<?php
namespace Ak\Lib\Type;

use BadMethodCallException;
use InvalidArgumentException;
use ReflectionClass;

/**
 * Implementacija enumeracije.
 *
 * @package Ak\Lib\Type
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/autokreso/lib
 *
 * @method static Enum DEFAULT()
 */
abstract class Enum
{
    const DEFAULT = 0;

    /** @var int|string */
    private $value;
    /** @var array */
    private static $constants = [];

    /**
     * Konstruktor.
     *
     * @param int|string $value
     */
    public function __construct($value = null)
    {
        $this->value = $value ?? static::DEFAULT;
        self::validate($this->value);
    }

    /**
     * Vraća postavljenu vrijednost.
     *
     * @return int|string
     */
    public function getValue()
    {
        return $this->value;
    }

    /**
     * Vraća postavljenu vrijednost kao string.
     */
    public function __toString()
    {
        return (string) $this->getValue();
    }

    /**
     * Vraća sve moguće vrijednosti ove enumeracije.
     *
     * @return int[]|string[]
     */
    public static function getValues(): array
    {
        $constants = self::getConstants();
        unset($constants['DEFAULT']);

        return array_values($constants);
    }

    /**
     * @param string $name
     * @param $arguments
     *
     * @return static
     */
    public static function __callStatic($name, $arguments)
    {
        $constants = self::getConstants();

        if (!array_key_exists($name, $constants)) {
            throw new BadMethodCallException(sprintf(
                "Enumeracija ne sadrži vrijednost s imenom '%s'.", $name
            ));
        }

        return new static($constants[$name]);
    }

    /**
     * Vraća konstante deklarirane u ovoj klasi.
     *
     * @return array
     */
    private static function getConstants(): array
    {
        if (!isset(self::$constants[static::class])) {
            $class = new ReflectionClass(static::class);
            $constants = $class->getConstants();

            unset($constants['DEFAULT']);

            self::$constants[static::class] = $constants;
        }

        return self::$constants[static::class];
    }

    /**
     * Provjerava da li je data vrijednost ispravna za ovu enumeraciju.
     *
     * @param int|string $value
     */
    private static function validate($value): void
    {
        if (!is_string($value) && !is_int($value)) {
            throw new InvalidArgumentException(sprintf(
                "Argument \$value mora biti string ili integer, a dobio sam %s.",
                is_object($value) ? get_class($value) : gettype($value)
            ));
        }

        $values = self::getValues();

        if (!in_array($value, $values)) {
            throw new InvalidArgumentException(sprintf(
                "Data vrijednost '%s' se ne nalazi u skupu { %s }.",
                (string) $value,
                implode(', ', $values)
            ));
        }
    }
}
